#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <string.h>
#include "matice_znaku.h"

#define NOP asm("nop")

//parametry obrazu
#define vga_pocet_radku 525 
#define znaku_na_radek 19   
#define radku_textu 20 
#define celkem_znaku 380        
#define vyska_znaku 24

#define vs_on   PORTD=0x00
#define vs_off  PORTD=0x40
#define hs_off PORTD |= (1 << PORTD7)
#define zapni_vykreslovani  DDRB=0xB0
#define vypni_vykreslovani  DDRB=0x90

volatile unsigned char pole_znaku[znaku_na_radek*radku_textu]; 
volatile unsigned int cislo_akt_znaku;
volatile unsigned char povol_vykreslovani;
volatile unsigned char akt_radek_textu;
volatile unsigned int radek;
volatile unsigned char akt_radek_znaku; 

//uvodni obrazovka po zapnuti, retezce ulozeny v programove pameti
const char str1[] PROGMEM =   "VGA zobrazovac";
const char str2[] PROGMEM =   "s mikrokontrolerem";
const char str3[] PROGMEM =   "ATMEL ATmega16";
const char str4[] PROGMEM =   "vytvorili:";
const char str5[] PROGMEM =   "Jan Polasek";
const char str6[] PROGMEM =   "Marek Dohnal";
const char str7[] PROGMEM =   "vedouci:";
const char str8[] PROGMEM =   "Ing. Zbynek Fedra";
const char str9[] PROGMEM =   "predmet MMIA";
const char str10[] PROGMEM =   "UREL FEKT VUT";
const char str11[] PROGMEM =   "1234567890123456789";
const char str12[] PROGMEM =   "   abcdefghijklm";
const char str13[] PROGMEM =   "   nopqrstuvwxyz";
const char str14[] PROGMEM =   "   ABCDEFGHIJKLM";
const char str15[] PROGMEM =   "   NOPQRSTUVWXYZ";

static void ini(void);

//synchronizacni rutina
ISR (TIMER0_OVF_vect)
{
	TCNT0 = 0xC3; 
	//pocitani radku, na konci snimku vynulovani promennych
	if (++radek == vga_pocet_radku)
	{
		radek = 0;
		akt_radek_textu = 0; 
		akt_radek_znaku = 0;
	}

	//VS dva radky, zaroven se nastavi pocatek HS
	if ((radek == 2 )||(radek == 3 ))
	{
		vs_on;
	}
	else
	{
		vs_off;
	}
		
	//povoleni vykreslovani v oblasti aktivniho videa
	//nez se toto provede, trva impuls HS
	if ((radek < 37)||(radek > 517))
	{
		povol_vykreslovani = 0;
	}
	else
	{
		povol_vykreslovani = 1;
		//pocitani radku textu
		if (++akt_radek_znaku == vyska_znaku)
		{
			akt_radek_textu++;
			akt_radek_znaku = 0;
		}
		else
		{
			NOP;
			NOP;
			NOP;
			NOP;
			NOP;
			NOP;
			NOP;
			NOP;
		}
		
	}
	hs_off;
}


//inicializace a nastaveni
static void ini(void)
{
	//nastaveni SPI
	DDRB=0xB0;
	SPSR = 1 << SPI2X;
	SPCR = (1 << SPE) | (1 << MSTR);

	//nastaveni UART, 8 datovych bitu, rychlost 1200 b/s
	UCSRB = 0x00;
	UCSRA = 0x00;
	UCSRC = 0x86;
	UBRRL = 0x40; 
	UBRRH = 0x03; 
	UCSRB = (1 << RXEN);

 	//nastaveni idle rezimu
 	set_sleep_mode(SLEEP_MODE_IDLE); 	
  
 	//nastaveni portu D pro synchronizaci
	DDRD = 0xC0;
	PORTD |= (1 << PORTD7);
	PORTD |= (1 << PORTD6);
	  
	//nastaveni casovace pro radkovou synchronizaci
	TCNT0 = 0xC3; 
	TCCR0 = 0x02;
	TIMSK = 0x01;
	//povoleni globalniho preruseni
	sei();
}


//prijem znaku a tvorba pole znaku pro vykreslovani
static void prijem(void)
{
	while(UCSRA & (1<<RXC))
	{
		unsigned char pom_radek, prijaty_znak;
		prijaty_znak = UDR;

		//kdyz je pole znaku plne, vynuluje se
		if(cislo_akt_znaku == (radku_textu*znaku_na_radek))
		{		
			short int i = celkem_znaku;
			char *ukazatel;
			ukazatel = &pole_znaku[0];
			while(i--)
			{
				*ukazatel++ = 0x0;
			};
			cislo_akt_znaku = 0x0;
		}
		switch ( prijaty_znak )
		{
			//osetreni Backspace
			case  8: 
				if(cislo_akt_znaku)
				{
					pole_znaku[cislo_akt_znaku] = 0x0;
					pole_znaku[--cislo_akt_znaku] = 0x0;
				}
			break;
					  
			//osetreni Enter
			case  13: 
				pom_radek = cislo_akt_znaku / znaku_na_radek;
				pole_znaku[cislo_akt_znaku] = 0x0;
				cislo_akt_znaku = pom_radek*znaku_na_radek + znaku_na_radek;
			break;
			//pokud neni prijaty znak Backspace ani Enter,
			//tak se prijaty znak zapise do pole znaku,
			//inkrementuje se cislo aktualniho znaku
			default: pole_znaku[cislo_akt_znaku++] = prijaty_znak;
		}
	}
}


//vykreslovani obrazu
void kresli_obraz()
{
	unsigned char i;
	char * uk1;
	const char * uk2;

	//naplneni pole znaku pro vykresleni uvodni obrazovky
	strcpy_P(&pole_znaku[znaku_na_radek*(radku_textu-20)],&str1[0]); 
	strcpy_P(&pole_znaku[znaku_na_radek*(radku_textu-19)],&str2[0]); 
	strcpy_P(&pole_znaku[znaku_na_radek*(radku_textu-18)],&str3[0]); 	
	strcpy_P(&pole_znaku[znaku_na_radek*(radku_textu-16)],&str4[0]);
	strcpy_P(&pole_znaku[znaku_na_radek*(radku_textu-15)],&str5[0]); 	
	strcpy_P(&pole_znaku[znaku_na_radek*(radku_textu-14)],&str6[0]);
	strcpy_P(&pole_znaku[znaku_na_radek*(radku_textu-12)],&str7[0]);
	strcpy_P(&pole_znaku[znaku_na_radek*(radku_textu-11)],&str8[0]);
	strcpy_P(&pole_znaku[znaku_na_radek*(radku_textu-9)],&str9[0]);
	strcpy_P(&pole_znaku[znaku_na_radek*(radku_textu-7)],&str10[0]);	
	strcpy_P(&pole_znaku[znaku_na_radek*(radku_textu-5)],&str11[0]);
	strcpy_P(&pole_znaku[znaku_na_radek*(radku_textu-4)],&str12[0]);
	strcpy_P(&pole_znaku[znaku_na_radek*(radku_textu-3)],&str13[0]);
	strcpy_P(&pole_znaku[znaku_na_radek*(radku_textu-2)],&str14[0]);
	strcpy_P(&pole_znaku[znaku_na_radek*(radku_textu-1)],&str15[0]);
	//cislo_akt_znaku se nastavi na konec pole znaku, aby pri
	//prijmu znaku pres USART doslo k vymazani obrazovky,
	//prijate znaky se zacnou vykreslovat od leveho
	//horniho rohu obrazovky
	cislo_akt_znaku = (radku_textu*znaku_na_radek);

	for(;;)
    {
		sleep_mode();
		
		
		
		//v oblasti aktivniho videa probiha vykreslovani obrazu
		//po dokonceni HS dojde k vykresleni jednoho radku obrazu
		if(povol_vykreslovani)
		{
			//do ukazatele 1 se ulozi adresa prvniho znaku
			//aktualniho radku textu v pole_znaku
			uk1 = &pole_znaku[akt_radek_textu * znaku_na_radek];

			//do ukazatele 2 se ulozi adresa aktualniho radku
			//znaku v matici znaku
			uk2 = &znak[0][akt_radek_znaku >> 1];
	
			//vykresleni jednoho VGA radku  
			i = znaku_na_radek;
			while(i--)
			{
				zapni_vykreslovani;
				//uk1 se kazdym pruchodem cyklu inkrementuje
				//a prechazi tak na dalsi znak na radku
				//kazdy radek matice znaku se kresli dvakrat
				SPDR = pgm_read_byte(uk2 + (* uk1++)*vyska_znaku/2);
			}
			//prodleva pro vykresleni posledniho znaku,
			//nez se vypne vykreslovani
			NOP;
			NOP;
			NOP;
			NOP;
			NOP;
			NOP;
			NOP;
			NOP;
			NOP;
			NOP;
			NOP;
			NOP;
			NOP;
		
			vypni_vykreslovani;
		}
		else
		{
			//kdyz je prijat znak, skok do funkce prijem
			if (UCSRA & (1<<RXC))
			{
				prijem();
			}

		}				
    }
}


int main(void)
{
    ini();
	kresli_obraz();
}
